/*
 * Library for Representing Binary Data as an Image
 * Copyright (c) 2020 b122251 <hi@b122251.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of mosquitto nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* File Inclusions */
#include <stdio.h>
#include <string.h>
#include "fixedint.h"
#include "draw.h"
#include "errors.h"

/* Definition of the order and nams of pixel formats */
const INTU16 pixelformats[NUMPIXELFORMATS] = {
	PIXELFORMAT_1BIT_GRAY,
	PIXELFORMAT_2BIT_GRAY,
	PIXELFORMAT_4BIT_GRAY,
	PIXELFORMAT_4BIT_GRAY_REVERSE,
	PIXELFORMAT_8BIT_GRAY,
	PIXELFORMAT_16BIT_XRGB1555,
	PIXELFORMAT_16BIT_RGB565,
	PIXELFORMAT_24BIT_RGB888,
	PIXELFORMAT_32BIT_XRGB8888
};
const char *pixelformatnames[NUMPIXELFORMATS] = {
	"1 Bit (B/W)",
	"2 Bit (Grayscale)",
	"4 Bit (Grayscale)",
	"4 Bit (Grayscale) (Reversed Nibbles)",
	"8 Bit (Grayscale)",
	"16 Bit (XRGB 1555)",
	"16 Bit (RGB 565)",
	"24 Bit (RGB 555)",
	"32 Bit (XRGB 8888)"
};

/* Pixel format conversion functions */
static void convertPixel_0001_0010(INTU8 *in, INTU16 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	if ((*in & ((INTU8) 0x01 << *in_curbit)) != (INTU8) 0x00)
	{
		*out = 0x7FFF;
	}
	else
	{
		*out = 0x0000;
	}
	
	if (*in_curbit == (INTU8) 0x00)
	{
		*in_curbit = (INTU8) 0x07;
		++(*in_curbyte);
	}
	else
	{
		--(*in_curbit);
	}
}

static void convertPixel_0001_0018(INTU8 *in, INTU8 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	if ((*in & ((INTU8) 0x01 << *in_curbit)) != (INTU8) 0x00)
	{
		out[0] = (INTU8) 0xFF;
		out[1] = (INTU8) 0xFF;
		out[2] = (INTU8) 0xFF;
	}
	else
	{
		out[0] = (INTU8) 0x00;
		out[1] = (INTU8) 0x00;
		out[2] = (INTU8) 0x00;
	}
	
	if (*in_curbit == (INTU8) 0x00)
	{
		*in_curbit = (INTU8) 0x07;
		++(*in_curbyte);
	}
	else
	{
		--(*in_curbit);
	}
}

static void convertPixel_0001_0020(INTU8 *in, INTU32 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	if ((*in & ((INTU8) 0x01 << *in_curbit)) != (INTU8) 0x00)
	{
		*out = 0x00FFFFFF;
	}
	else
	{
		*out = 0x00000000;
	}
	
	if (*in_curbit == (INTU8) 0x00)
	{
		*in_curbit = (INTU8) 0x07;
		++(*in_curbyte);
	}
	else
	{
		--(*in_curbit);
	}
}

static void convertPixel_0002_0010(INTU8 *in, INTU16 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	INTU8 shiftval = (*in_curbit & (INTU8) 0xFE);

	*out = (INTU16) (*in & ((INTU8) 0x03 << shiftval));
	*out >>= shiftval;
	*out <<= 3U;
	*out |= (*out >> 2U);
	*out |= (*out >> 4U);
	*out |= (*out << 10U);
	*out |= (*out >> 5U);

	if (*in_curbit < (INTU8) 0x02)
	{
		*in_curbit += (INTU8) 0x08;
		++(*in_curbyte);
	}
	*in_curbit -= (INTU8) 0x02;
}

static void convertPixel_0002_0018(INTU8 *in, INTU8 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	INTU8 shiftval = (*in_curbit & (INTU8) 0xFE);

	out[0] = (INTU8) (*in & ((INTU8) 0x03 << shiftval));
	out[0] >>= shiftval;
	out[0] |= (out[0] << 4U);
	out[0] |= (out[0] << 2U);
	out[1] = out[0];
	out[2] = out[0];

	if (*in_curbit < (INTU8) 0x02)
	{
		*in_curbit += (INTU8) 0x08;
		++(*in_curbyte);
	}
	*in_curbit -= (INTU8) 0x02;
}

static void convertPixel_0002_0020(INTU8 *in, INTU32 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	INTU8 shiftval = (*in_curbit & (INTU8) 0xFE);

	*out = (INTU32) (*in & ((INTU8) 0x03 << shiftval));
	*out >>= shiftval;
	*out |= (*out << 4U);
	*out |= (*out << 2U);
	*out |= (*out << 8U);
	*out |= (*out << 8U);

	if (*in_curbit < (INTU8) 0x02)
	{
		*in_curbit += (INTU8) 0x08;
		++(*in_curbyte);
	}
	*in_curbit -= (INTU8) 0x02;
}

static void convertPixel_0004_0010(INTU8 *in, INTU16 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	INTU8 shiftval = (INTU8) (*in_curbit & (INTU8) 0xFC);

	*out = (INTU16) (*in & ((INTU8) 0x0F << shiftval));
	*out >>= shiftval;
	*out <<= 1U;
	*out |= (*out >> 4U);
	*out |= (*out << 10U);
	*out |= (*out >> 5U);

	if (*in_curbit < (INTU8) 0x04)
	{
		*in_curbit += (INTU8) 0x08;
		++(*in_curbyte);
	}
	*in_curbit -= (INTU8) 0x04;
}

static void convertPixel_0004_0018(INTU8 *in, INTU8 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	INTU8 shiftval = (INTU8) (*in_curbit & (INTU8) 0xFC);
	
	out[0] = (INTU8) (*in & ((INTU8) 0x0F << shiftval));
	out[0] >>= shiftval;
	out[0] |= (out[0] << 4U);
	out[1] = out[0];
	out[2] = out[0];

	if (*in_curbit < (INTU8) 0x04)
	{
		*in_curbit += (INTU8) 0x08;
		++(*in_curbyte);
	}
	*in_curbit -= (INTU8) 0x04;
}

static void convertPixel_0004_0020(INTU8 *in, INTU32 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	INTU8 shiftval = (INTU8) (*in_curbit & (INTU8) 0xFC);

	*out = (INTU32) (*in & ((INTU8) 0x0F << shiftval));
	*out >>= shiftval;
	*out |= (*out << 4U);
	*out |= (*out << 8U);
	*out |= (*out << 8U);

	if (*in_curbit < (INTU8) 0x04)
	{
		*in_curbit += (INTU8) 0x08;
		++(*in_curbyte);
	}
	*in_curbit -= (INTU8) 0x04;
}

static void convertPixel_1004_0010(INTU8 *in, INTU16 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	INTU8 shiftval = (INTU8) (*in_curbit & (INTU8) 0xFC);

	*out = (INTU16) (*in & ((INTU8) 0xF0 >> shiftval));
	*out <<= shiftval;
	*out <<= 1U;
	*out |= (*out >> 4U);
	*out |= (*out << 10U);
	*out |= (*out >> 5U);

	if (*in_curbit < (INTU8) 0x04)
	{
		*in_curbit += (INTU8) 0x08;
		++(*in_curbyte);
	}
	*in_curbit -= (INTU8) 0x04;
}

static void convertPixel_1004_0018(INTU8 *in, INTU8 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	INTU8 shiftval = (INTU8) (*in_curbit & (INTU8) 0xFC);
	
	out[0] = (INTU8) (*in & ((INTU8) 0xF0 >> shiftval));
	out[0] <<= shiftval;
	out[0] |= (out[0] << 4U);
	out[1] = out[0];
	out[2] = out[0];

	if (*in_curbit < (INTU8) 0x04)
	{
		*in_curbit += (INTU8) 0x08;
		++(*in_curbyte);
	}
	*in_curbit -= (INTU8) 0x04;
}

static void convertPixel_1004_0020(INTU8 *in, INTU32 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	INTU8 shiftval = (INTU8) (*in_curbit & (INTU8) 0xFC);

	*out = (INTU32) (*in & ((INTU8) 0xF0 >> shiftval));
	*out <<= shiftval;
	*out |= (*out << 4U);
	*out |= (*out << 8U);
	*out |= (*out << 8U);

	if (*in_curbit < (INTU8) 0x04)
	{
		*in_curbit += (INTU8) 0x08;
		++(*in_curbyte);
	}
	*in_curbit -= (INTU8) 0x04;
}

static void convertPixel_0008_0010(INTU8 *in, INTU16 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	*out = (INTU16) *in;
	*out >>= 3U;
	*out |= (*out << 10U);
	*out |= (*out >> 5U);
	
	++(*in_curbyte);
}

static void convertPixel_0008_0018(INTU8 *in, INTU8 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	out[0] = *in;
	out[1] = out[0];
	out[2] = out[0];
	
	++(*in_curbyte);
}

static void convertPixel_0008_0020(INTU8 *in, INTU32 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	*out = (INTU32) *in;
	*out |= (*out << 8U);
	*out |= (*out << 8U);
	
	++(*in_curbyte);
}

static void convertPixel_0010_0010(INTU16 *in, INTU16 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	*out = *in;
	*in_curbyte += 0x00000002;
}

static void convertPixel_0010_0018(INTU16 *in, INTU8 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	*out = *in;
	out[0] = (INTU8) ((*in & 0x7C00) >> 7U);
	out[1] = (INTU8) ((*in & 0x03E0) >> 2U);
	out[2] = (INTU8) ((*in & 0x001F) << 3U);
	out[0] |= (out[0] >> 5U);
	out[1] |= (out[1] >> 5U);
	out[2] |= (out[2] >> 5U);
	*in_curbyte += 0x00000002;
}

static void convertPixel_0010_0020(INTU16 *in, INTU32 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	*out = (INTU32) ((*in & 0x7C00) >> 7U);
	*out <<= 8U;
	*out |= (INTU32) ((*in & 0x03E0) >> 2U);
	*out <<= 8U;
	*out |= (INTU32) ((*in & 0x001F) << 3U);
	*out |= ((*out & 0x00E0E0E0) >> 5U);
	*in_curbyte += 0x00000002;
}

static void convertPixel_0110_0010(INTU16 *in, INTU16 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	*out = ((*in & 0xFFC0) >> 1U);
	*out |= (*in & 0x001F);
	*in_curbyte += 0x00000002;
}

static void convertPixel_0110_0018(INTU16 *in, INTU8 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	*out = *in;
	out[0] = (INTU8) ((*in & 0xF800) >> 8U);
	out[1] = (INTU8) ((*in & 0x07E0) >> 3U);
	out[2] = (INTU8) ((*in & 0x001F) << 3U);
	out[0] |= (out[0] >> 5U);
	out[1] |= (out[1] >> 6U);
	out[2] |= (out[2] >> 5U);
	*in_curbyte += 0x00000002;
}

static void convertPixel_0110_0020(INTU16 *in, INTU32 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	*out = (INTU32) ((*in & 0xF800) >> 8U);
	*out <<= 8U;
	*out |= (INTU32) ((*in & 0x07E0) >> 3U);
	*out <<= 8U;
	*out |= (INTU32) ((*in & 0x001F) << 3U);
	*out |= ((*out & 0x00E000E0) >> 5U);
	*out |= ((*out & 0x0000C000) >> 6U);
	*in_curbyte += 0x00000002;
}

static void convertPixel_0018_0010(INTU8 *in, INTU16 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	*out = (in[0] >> 2U);
	*out <<= 5U;
	*out |= (in[1] >> 2U);
	*out <<= 5U;
	*out |= (in[2] >> 2U);
	*out <<= 5U;
	*in_curbyte += 0x00000003;
}

static void convertPixel_0018_0018(INTU8 *in, INTU8 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	out[0] = in[0];
	out[1] = in[1];
	out[2] = in[2];
	*in_curbyte += 0x00000003;
}

static void convertPixel_0018_0020(INTU8 *in, INTU32 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	*out = (INTU32) in[0];
	*out <<= 8U;
	*out |= (INTU32) in[1];
	*out <<= 8U;
	*out |= (INTU32) in[2];
	*in_curbyte += 0x00000003;
}

static void convertPixel_0020_0010(INTU32 *in, INTU16 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	*out = (INTU16) ((*in & 0x00F80000) >> 9U);
	*out |= (INTU16) ((*in & 0x0000F800) >> 6U);
	*out |= (INTU16) ((*in & 0x000000F8) >> 3U);

	*in_curbyte += 0x00000004;
}

static void convertPixel_0020_0018(INTU32 *in, INTU8 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	out[0] = (INTU8) ((*in & 0x00FF0000) >> 16U);
	out[1] = (INTU8) ((*in & 0x0000FF00) >> 8U);
	out[2] = (INTU8) (*in & 0x000000FF);

	*in_curbyte += 0x00000004;
}

static void convertPixel_0020_0020(INTU32 *in, INTU32 *out,
                                   INTU32 *in_curbyte, INTU8 *in_curbit)
{
	*out = *in;

	*in_curbyte += 0x00000004;
}

/* Pixel Doubling Functions */
static void doublePixel16(INTU16 *in)
{
	in[1] = in[0];
}

static void doublePixel24(INTU8 *in)
{
	in[3] = in[0];
	in[4] = in[1];
	in[5] = in[2];
}

static void doublePixel32(INTU32 *in)
{
	in[1] = in[0];
}

/*
 * Function that draws the actual screen.
 * Parameters:
 * * file = In-memory copy of the file
 * * filesize = Size of file (in bytes)
 * * offset = From what offset of the file to start reading
 * * offset_b = Number of bits beyond the offset
 * * width = How wide the image is displayed
 * * informat = Input pixel format
 * * output = Pointer to the output window
 * * outwidth = Width of the output window
 * * outheight = Height of the output window
 * * outformat = Output pixel format
 * * scalefactor = How many times to print a pixel
 */
int drawscreen(INTU8 *file, INTU32 filesize, INTU32 offset, INTU8 offset_b,
               INTU32 width, INTU16 informat, INTU8 *output, INTU32 outwidth,
               INTU32 outheight, INTU16 outformat, INTU32 scalefactor)
{
	/* Variable Declarations */
	INTU32 in_curbyte;     /* Current byte in the input file */
	INTU8  in_curbit;      /* Current bit in the input file */
	INTU32 in_linestart_B; /* Byte in the input where the line started */
	INTU8  in_linestart_b; /* Value of in_curbit when the line started */
	INTU32 in_linesize_B;  /* Number of bytes a line is long */
	INTU8  in_linesize_b;  /* Number of bits a line is longer than bytes */
	INTU32 in_numLines;    /* Number of lines of pixels the file is long */
	INTU32 in_restPixels;  /* Number of pixels to draw on the final line */
	INTU32 out_linestart;  /* Position in output where current line started */
	INTU32 out_lineend;    /* Position in output where current line ends */
	INTU32 out_curpos;     /* Current position in output */
	INTU32 out_drawheight; /* Number of unique lines to be drawn */
	INTU32 out_drawwidth;  /* Number of unique pixels per line to be drawn */
	INTU32 out_Bpp;        /* Number of bytes per pixel in the output */
	INTU32 out_linesize;   /* The size of a line of pixels in output */
	INTU32 out_drawnwsize; /* The size of the width drawn (in bytes) */
	INTU32 out_curline;    /* Current unique line in the output */
	INTU32 out_curduplic;  /* Current duplicate based on scalefactor */
	/* Pointer to the function that converts the pixel from in to out format */
	void (*convertPixel)(void *, void *, INTU32 *, INTU8 *) = NULL;
	/* Pointer to the function that duplicates the output pixel */
	void (*doublePixel)(void *) = NULL;

	/* Initialise input variables */
	in_linesize_B = (INTU32) (informat & 0x00FF);
	in_linesize_B *= width;
	in_linesize_b = (INTU8) (in_linesize_B & 0x00000007);
	in_linesize_B >>= 3U;
	in_linestart_B = offset;
	in_linestart_b = offset_b;

	/* Initialise output variables */
	out_Bpp = (INTU32) ((outformat & 0x00FF) >> 3U);
	out_drawwidth = (outwidth / scalefactor);
	if (out_drawwidth > width)
	{
		out_drawwidth = width;
	}
	out_drawheight = (outheight / scalefactor);
	out_linesize = (outwidth * out_Bpp);
	out_drawnwsize = (out_drawwidth * scalefactor * out_Bpp);

	/* Clears the screen */
	memset(output, 0, (size_t) (outwidth * outheight * out_Bpp));

	/* Ensures drawing is stopped when the end of the file is reached */
	if (offset >= filesize)
	{
		return ERROR_NONE;
	}
	in_numLines = (filesize - offset);
	in_numLines <<= 3U;
	in_numLines /= (INTU32) (informat & 0x00FF);
	in_restPixels = (in_numLines % width);
	in_numLines /= width;
	if (out_drawheight > in_numLines)
	{
		out_drawheight = in_numLines;
	}
	else
	{
		in_restPixels = 0x00000000;
	}
	if (in_restPixels > out_drawwidth)
	{
		in_restPixels = out_drawwidth;
	}
	
	/* Sets pixel conversion function */
	switch (outformat)
	{
		case PIXELFORMAT_16BIT_XRGB1555:
			switch (informat)
			{
				case PIXELFORMAT_1BIT_GRAY:
					convertPixel = (void *) &convertPixel_0001_0010;
					break;
				case PIXELFORMAT_2BIT_GRAY:
					convertPixel = (void *) &convertPixel_0002_0010;
					break;
				case PIXELFORMAT_4BIT_GRAY:
					convertPixel = (void *) &convertPixel_0004_0010;
					break;
				case PIXELFORMAT_4BIT_GRAY_REVERSE:
					convertPixel = (void *) &convertPixel_1004_0010;
					break;
				case PIXELFORMAT_8BIT_GRAY:
					convertPixel = (void *) &convertPixel_0008_0010;
					break;
				case PIXELFORMAT_16BIT_XRGB1555:
					convertPixel = (void *) &convertPixel_0010_0010;
					break;
				case PIXELFORMAT_16BIT_RGB565:
					convertPixel = (void *) &convertPixel_0110_0010;
					break;
				case PIXELFORMAT_24BIT_RGB888:
					convertPixel = (void *) &convertPixel_0018_0010;
					break;
				case PIXELFORMAT_32BIT_XRGB8888:
					convertPixel = (void *) &convertPixel_0020_0010;
					break;
			}
			break;
		case PIXELFORMAT_24BIT_RGB888:
			switch (informat)
			{
				case PIXELFORMAT_1BIT_GRAY:
					convertPixel = (void *) &convertPixel_0001_0018;
					break;
				case PIXELFORMAT_2BIT_GRAY:
					convertPixel = (void *) &convertPixel_0002_0018;
					break;
				case PIXELFORMAT_4BIT_GRAY:
					convertPixel = (void *) &convertPixel_0004_0018;
					break;
				case PIXELFORMAT_4BIT_GRAY_REVERSE:
					convertPixel = (void *) &convertPixel_1004_0018;
					break;
				case PIXELFORMAT_8BIT_GRAY:
					convertPixel = (void *) &convertPixel_0008_0018;
					break;
				case PIXELFORMAT_16BIT_XRGB1555:
					convertPixel = (void *) &convertPixel_0010_0018;
					break;
				case PIXELFORMAT_16BIT_RGB565:
					convertPixel = (void *) &convertPixel_0110_0018;
					break;
				case PIXELFORMAT_24BIT_RGB888:
					convertPixel = (void *) &convertPixel_0018_0018;
					break;
				case PIXELFORMAT_32BIT_XRGB8888:
					convertPixel = (void *) &convertPixel_0020_0018;
					break;
			}
			break;
		case PIXELFORMAT_32BIT_XRGB8888:
			switch (informat)
			{
				case PIXELFORMAT_1BIT_GRAY:
					convertPixel = (void *) &convertPixel_0001_0020;
					break;
				case PIXELFORMAT_2BIT_GRAY:
					convertPixel = (void *) &convertPixel_0002_0020;
					break;
				case PIXELFORMAT_4BIT_GRAY:
					convertPixel = (void *) &convertPixel_0004_0020;
					break;
				case PIXELFORMAT_4BIT_GRAY_REVERSE:
					convertPixel = (void *) &convertPixel_1004_0020;
					break;
				case PIXELFORMAT_8BIT_GRAY:
					convertPixel = (void *) &convertPixel_0008_0020;
					break;
				case PIXELFORMAT_16BIT_XRGB1555:
					convertPixel = (void *) &convertPixel_0010_0020;
					break;
				case PIXELFORMAT_16BIT_RGB565:
					convertPixel = (void *) &convertPixel_0110_0020;
					break;
				case PIXELFORMAT_24BIT_RGB888:
					convertPixel = (void *) &convertPixel_0018_0020;
					break;
				case PIXELFORMAT_32BIT_XRGB8888:
					convertPixel = (void *) &convertPixel_0020_0020;
					break;
			}
			break;
	}

	/* Sets doublePixel */
	if ((outformat & 0x00FF) == 0x0010)
	{
		doublePixel = (void *) &doublePixel16;
	}
	else if ((outformat & 0x00FF) == 0x0018)
	{
		doublePixel = (void *) &doublePixel24;
	}
	else if ((outformat & 0x00FF) == 0x0020)
	{
		doublePixel = (void *) &doublePixel32;
	}

	/* Checks that convertPixel and doublePixel are set */
	if (convertPixel == NULL)
	{
		return ERROR_NO_CONVERTPIXEL_FUNCTION;
	}
	if (doublePixel == NULL)
	{
		return ERROR_NO_DOUBLEPIXEL_FUNCTION;
	}

	/* Draws every unique line */
	out_linestart = 0x00000000;
	for (out_curline = 0x00000000; out_curline < out_drawheight; ++out_curline)
	{
		/* Fills out the current line pixel by pixel */
		in_curbyte = in_linestart_B;
		in_curbit = in_linestart_b;
		out_lineend = (out_linestart + out_drawnwsize);
		out_curpos = out_linestart;
		while (out_curpos < out_lineend)
		{
			/* Calculates the current pixel */
			(*convertPixel)(&file[in_curbyte], &output[out_curpos],
			                &in_curbyte, &in_curbit);

			/* Duplicates the current pixel if scalefactor > 1 */
			for (out_curduplic = 0x00000001; out_curduplic < scalefactor;
			     ++out_curduplic)
			{
				(*doublePixel)(&output[out_curpos]);
				out_curpos += out_Bpp;
			}
			out_curpos += out_Bpp;
		}
		
		/* Duplictes the line if scalefactor > 1 */
		for (out_curduplic = 0x00000001; out_curduplic < scalefactor;
		     ++out_curduplic)
		{
			out_curpos = (out_linestart + out_linesize);
			memcpy(&output[out_curpos], &output[out_linestart],
			       (size_t) out_drawnwsize);
			out_linestart += out_linesize;
		}

		/* Moves on to the next line in the input */
		in_linestart_B += in_linesize_B;
		/* Adjusts the linestart in bits if needed */
		if (in_linesize_b != (INTU8) 0x00)
		{
			/* Wraps around if this crosses the edge of a byte */
			if (in_linestart_b < in_linesize_b)
			{
				++in_linestart_B;
				in_linestart_b += (INTU8) 0x08;
			}
			in_linestart_b -= in_linesize_b;
		}
		
		/* Moves on to the next line in the output */
		out_linestart += out_linesize;
	}

	/* Draws the remaining pixels if needed */
	if (in_restPixels > 0x00000000)
	{
		in_curbyte = in_linestart_B;
		in_curbit = in_linestart_b;
		out_lineend = (out_linestart + out_drawnwsize);
		out_curpos = out_linestart;
		
		for (; in_restPixels > 0x00000000; --in_restPixels)
		{
			/* Calculates the current pixel */
			(*convertPixel)(&file[in_curbyte], &output[out_curpos],
			                &in_curbyte, &in_curbit);
			/* Duplicates the current pixel if scalefactor > 1 */
			for (out_curduplic = 0x00000001; out_curduplic < scalefactor;
			     ++out_curduplic)
			{
				(*doublePixel)(&output[out_curpos]);
				out_curpos += out_Bpp;
			}
			out_curpos += out_Bpp;
		}
		
		/* Duplictes the line if scalefactor > 1 */
		for (out_curduplic = 0x00000001; out_curduplic < scalefactor;
		     ++out_curduplic)
		{
			out_curpos = (out_linestart + out_linesize);
			memcpy(&output[out_curpos], &output[out_linestart],
			       (size_t) out_drawnwsize);
			out_linestart += out_linesize;
		}
	}

	return ERROR_NONE;
}
