/*
 * Library that Handles Keyboard Input (POSIX-Version)
 * Copyright (c) 2020 b122251 <hi@b122251.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of mosquitto nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* File Inclusions */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <termios.h>
#include <unistd.h>
#include "fixedint.h"
#include "input.h"
#include "draw.h"
#include "errors.h"
#include "version.h"

/* Macro's for standard I/O */
#define STDIN  0
#define STDOUT 1

/* Special characters for UI */
#define CHAR_INVALID      (char) 0x00
#define CHAR_NEWLINE      (char) 0x0A
#define CHAR_CURSOR_UP    (char) 0x80
#define CHAR_CURSOR_DOWN  (char) 0x81
#define CHAR_CURSOR_LEFT  (char) 0x82
#define CHAR_CURSOR_RIGHT (char) 0x83
#define CHAR_ESC          (char) 0x84
#define CHAR_BACKSPACE    (char) 0x85
#define CHAR_PGUP         (char) 0x86
#define CHAR_PGDOWN       (char) 0x87
#define CHAR_F1           (char) 0x88

/* Terminal settings */
static struct termios term;      /* Terminal settings */
static struct termios term_orig; /* Original terminal settings */

/* Buffer for textual input */
char inbuffer[INBUFSIZE];
INTU16 inbufferpos;

/* Internal Function Declarations */
static void processchar(char c, INTU32 *offset, INTU8 *offset_b, INTU32 *width,
                        INTU16 *format, INTU32 *scalefactor, int *progstate,
                        INTU8 *redraw, INTU8 *end_program, INTU8 *hide_hud,
                        INTU8 *hud_hex, INTU32 filesize);

/*
 * Function that initialises the terminal settings
 */
int terminal_init(void)
{
	/* Variable Declarations */
	int funcval = 0; /* Return values for called functions */
	char cursoff[12] = {0x1B, 0x5B, 0x3F,
	                    0x32, 0x35, 0x6C, 
	                    0x1B, 0x5B, 0x3F,
	                    0x31, 0x63, 0x00}; /* Sets cursor off */

	/* Gets the original terminal settings */
	funcval = tcgetattr(STDIN, &term_orig);
	if (funcval != 0)
	{
		printf(PROGNAME ": tcgetattr() failed\n");
		return ERROR_OSDEP;
	}

	/* Makes modified terminal settings */
	memcpy(&term, &term_orig, sizeof(term));
	term.c_lflag &= ~ICANON;
	term.c_lflag &= ~ECHO;
	term.c_cc[VMIN] = 0;
	term.c_cc[VTIME] = 0;

	/* Sets it */
	funcval = tcsetattr(STDIN, TCSANOW, &term);
	if (funcval != 0)
	{
		printf(PROGNAME ": tcsetattr() failed\n");
		return ERROR_OSDEP;
	}

	/* Turns off the flickering cursor */
    funcval = write(STDOUT, &cursoff, 11);
	if (funcval != 11)
	{
		printf(PROGNAME ": write() failed on STDOUT\n");
		return ERROR_OSDEP;
	}

	return ERROR_NONE;
}

/*
 * Function that resets the original terminal settings
 */
int terminal_uninit(void)
{
	/* Variable Declarations */
	int funcval = 0; /* Return values for called functions */
	char curson[12] = {0x1B, 0x5B, 0x3F,
	                   0x32, 0x35, 0x68,
	                   0x1B, 0x5B, 0x3F,
	                   0x30, 0x63, 0x00};
	
	/* Sets it */
	funcval = tcsetattr(0, TCSANOW, &term_orig);
	if (funcval != 0)
	{
		printf(PROGNAME ": tcsetattr() failed\n");
		return ERROR_OSDEP;
	}

	/* Turns on the flickering cursor */
    funcval = write(STDOUT, &curson, 11);
	if (funcval != 11)
	{
		printf(PROGNAME ": write() failed on STDOUT\n");
		return ERROR_OSDEP;
	}

	return ERROR_NONE;
}

/*
 * Thread that continually handles input.
 */
void inputThread(void *in)
{
	/* Variable Declarations */
	struct input_thread_parameters *i; /* Parameters */
	fd_set rcvfds;          /* Set of file descriptors for stdin */
	int selectval;          /* Return value for select() */
	int readval;            /* Return value for read() */
	char c[4];              /* Character to be read */

	/* Gets the input parameters */
	i = (struct input_thread_parameters *) in;

	while (0 == 0)
	{
		/* Ends this thread if the other one has crashed */
		if (*(i->end_program) == (INTU8) 0x01)
		{
			break;
		}
		FD_ZERO(&rcvfds);
		FD_SET(STDIN, &rcvfds);
		selectval = select((STDIN + 1), &rcvfds, NULL, NULL, NULL);
		if (selectval > 0)
		{
			readval = read(STDIN, c, 4);
			if (readval < 0)
			{
				printf(PROGNAME ": Failed to read from stdin\n");
				*(i->retval) = ERROR_OSDEP;
				break;
			}
			else if (readval == 4)
			{
				if ((c[0] == (char) 0x1B) &&
				    (c[1] == (char) 0x5B) &&
				    (c[2] == (char) 0x35) &&
					(c[3] == (char) 0x7E))
				{
					c[0] = CHAR_PGUP;
				}
				else if ((c[0] == (char) 0x1B) &&
				         (c[1] == (char) 0x5B) &&
				         (c[2] == (char) 0x36) &&
					     (c[3] == (char) 0x7E))
				{
					c[0] = CHAR_PGDOWN;
				}
				else
				{
					c[0] = CHAR_INVALID;
				}
			}
			else if (readval == 3)
			{
				if ((c[0] == (char) 0x1B) && /* ESC */
				    (c[1] == (char) 0x5B) && /* [ */
				    (c[2] == 'A'))
				{
					c[0] = CHAR_CURSOR_UP;
				}
				else if ((c[0] == (char) 0x1B) && /* ESC */
				         (c[1] == (char) 0x5B) && /* [ */
				         (c[2] == 'B'))
				{
					c[0] = CHAR_CURSOR_DOWN;
				}
				else if ((c[0] == (char) 0x1B) && /* ESC */
				         (c[1] == (char) 0x5B) && /* [ */
				         (c[2] == 'C'))
				{
					c[0] = CHAR_CURSOR_RIGHT;
				}
				else if ((c[0] == (char) 0x1B) && /* ESC */
				         (c[1] == (char) 0x5B) && /* [ */
				         (c[2] == 'D'))
				{
					c[0] = CHAR_CURSOR_LEFT;
				}
				else if ((c[0] == (char) 0x1B) && /* ESC */
				         (c[1] == (char) 0x4F) && /* [ */
				         (c[2] == (char) 0x50))
				{
					c[0] = CHAR_F1;
				}

				else
				{
					c[0] = CHAR_INVALID;
				}
			}
			else if (readval == 1)
			{
				if ((c[0] == (char) 0x7F) || (c[0] == (char) 0x08))
				{
					c[0] = CHAR_BACKSPACE;
				}
				else if (c[0] == (char) 0x1B)
				{
					c[0] = CHAR_ESC;
				}
				else if (c[0] == '\n')
				{
					c[0] = CHAR_NEWLINE;
				}
				else if (c[0] < ' ')
				{
					c[0] = CHAR_INVALID;
				}
			}

			/* Processes the character */
			processchar(c[0], i->offset, i->offset_b, i->width, i->format,
			            i->scalefactor, i->progstate, i->redraw,
			            i->end_program, i->hide_hud, i->hud_hex, i->filesize);
		}
	}

	i->retval = ERROR_NONE;
}

/*
 * Function that processes the entered character
 */
static void processchar(char c, INTU32 *offset, INTU8 *offset_b, INTU32 *width,
                        INTU16 *format, INTU32 *scalefactor, int *progstate,
                        INTU8 *redraw, INTU8 *end_program, INTU8 *hide_hud,
                        INTU8 *hud_hex, INTU32 filesize)
{
	/* Variable Declarations */
	INTU32 widthBytes; /* Number of bytes for the width */
	INTU32 maxWidth;   /* Maximum allowable width value */
	INTU32 temp;       /* Temporary value */
	INTU8  widthBits;  /* Numeber of bits after the bytes */
	int    operation;  /* What to do with the number (0 set, 1 add, 2 sub) */
	int    curpos;     /* Current position in the buffer */
	int    temphex;    /* Whether the entered value is hexadecimal */

	if (*progstate == PROGSTATE_NORMAL)
	{
		switch (c)
		{
			/* Go up one row of pixels */
			case CHAR_CURSOR_UP:
			case 'k':
				widthBytes = (INTU32) (pixelformats[*format] & 0x00FF);
				widthBytes *= *width;
				widthBits = (INTU8) (widthBytes & 0x00000007);
				widthBytes >>= 3U;
				if (*offset > widthBytes)
				{
					*offset -= widthBytes;
					*offset_b += widthBits;
					if (*offset_b >= (INTU8) 0x08)
					{
						--(*offset);
						*offset_b -= (INTU8) 0x08;
					}
				}
				else if (*offset > 0x00000000)
				{
					*offset = 0x00000000;
					*offset_b = (INTU8) 0x07;
				}
				(*redraw)++;
				break;
			/* Go down one row of pixels */
			case CHAR_CURSOR_DOWN:
			case 'j':
				widthBytes = (INTU32) (pixelformats[*format] & 0x00FF);
				widthBytes *= *width;
				widthBits = (INTU8) (widthBytes & 0x00000007);
				widthBytes >>= 3U;
				*offset += widthBytes;
				if (*offset_b < widthBits)
				{
					*offset_b += (INTU8) 0x08;
					++(*offset);
				}
				*offset_b -= widthBits;
				(*redraw)++;
				break;
			/* Go one pixel to the left */
			case CHAR_CURSOR_LEFT:
				widthBytes = (INTU32) (pixelformats[*format] & 0x00FF);
				widthBits = (INTU8) (widthBytes & 0x00000007);
				widthBytes >>= 3U;
				if (*offset > widthBytes)
				{
					*offset -= widthBytes;
					*offset_b += widthBits;
					if (*offset_b >= (INTU8) 0x08)
					{
						--(*offset);
						*offset_b -= (INTU8) 0x08;
					}
				}
				else
				{
					*offset = 0x00000000;
					if (*offset_b != (INTU8) 0x07)
					{
						*offset_b += widthBits;
						if (*offset_b >= (INTU8) 0x08)
						{
							*offset_b = (INTU8) 0x07;
						}
					}
				}
				(*redraw)++;
				break;
			/* Go one pixel to the right */
			case CHAR_CURSOR_RIGHT:
				widthBytes = (INTU32) (pixelformats[*format] & 0x00FF);
				widthBits = (INTU8) (widthBytes & 0x00000007);
				widthBytes >>= 3U;
				*offset += widthBytes;
				if (*offset_b < widthBits)
				{
					*offset_b += (INTU8) 0x08;
					++(*offset);
				}
				*offset_b -= widthBits;
				(*redraw)++;
				break;
			/* Go one byte to the left */
			case 'h':
				if (*offset > 0x00000000)
				{
					--(*offset);
				}
				else
				{
					*offset_b = (INTU8) 0x07;
				}
				(*redraw)++;
				break;
			/* Go one byte to the right */
			case 'l':
				++(*offset);
				(*redraw)++;
				break;
			/* Go up 16 lines */
			case 'K':
			case CHAR_PGUP:
				widthBytes = (INTU32) (pixelformats[*format] & 0x00FF);
				widthBytes *= *width;
				widthBytes <<= 1U;
				if (*offset > widthBytes)
				{
					*offset -= widthBytes;
				}
				else if (*offset > 0x00000000)
				{
					*offset = 0x00000000;
					*offset_b = (INTU8) 0x07;
				}
				(*redraw)++;

				break;
			/* Go down 16 lines */
			case 'J':
			case CHAR_PGDOWN:
				widthBytes = (INTU32) (pixelformats[*format] & 0x00FF);
				widthBytes *= *width;
				widthBytes <<= 1U;
				*offset += widthBytes;
				(*redraw)++;
				break;
			/* Subtract one pixel from width */
			case 'H':
				if (*width > 0x00000001)
				{
					(*width)--;
					(*redraw)++;
				}
				break;
			/* Add one pixel to width */
			case 'L':
				(*width)++;
				if (*width == 0x00000000)
				{
					*width = 0x00000001;
				}
				(*redraw)++;
				break;
			/* Zoom in */
			case '+':
				(*scalefactor)++;
				if (*scalefactor == 0x00000000)
				{
					*scalefactor = 0x00000001;
				}
				(*redraw)++;
				break;
			/* Zoom out */
			case '-':
				if (*scalefactor > 0x00000001)
				{
					(*scalefactor)--;
					(*redraw)++;
				}
				break;
			/* Reset zoomlevel */
			case '0':
				if (*scalefactor != 0x00000001)
				{
					*scalefactor = 0x00000001;
					(*redraw)++;
				}
				break;
			/* Go to the next pixel format */
			case 'p':
				(*format)++;
				if (*format >= NUMPIXELFORMATS)
				{
					*format = 0x0000;
				}
				*offset_b = (INTU8) 0x07;
				(*redraw)++;
				break;
			/* Go to the previous pixel format */
			case 'P':
				if (*format == 0x0000)
				{
					*format += NUMPIXELFORMATS;
				}
				--(*format);
				*offset_b = (INTU8) 0x07;
				(*redraw)++;
				break;
			/* Display the help screen */
			case '?':
			case CHAR_F1:
				*progstate = PROGSTATE_HELPSCREEN;
				(*redraw)++;
				break;
			/* End the program */
			case 'q':
			case 'Q':
			case CHAR_ESC:
				*end_program = (INTU8) 0x01;
				break;
			/* Enter width */
			case 'w':
			case 'W':
				*progstate = PROGSTATE_SET_WIDTH;
				(*redraw)++;
				break;
			/* Enter offset */
			case 'o':
			case 'O':
				*progstate = PROGSTATE_SET_OFFSET;
				(*redraw)++;
				break;
			/* Toggle hexadecimal display */
			case 'x':
			case 'X':
				(*hud_hex) ^= (INTU8) 0x01;
				(*redraw)++;
				break;
			/* Toggle display */
			case 'v':
			case 'V':
				(*hide_hud) ^= (INTU8) 0x01;
				(*redraw)++;
				break;
		}
	}
	else if (*progstate == PROGSTATE_HELPSCREEN)
	{
		/* Exits the help screen when any key is pressed */
		*progstate = PROGSTATE_NORMAL;
		(*redraw)++;
		if ((c != CHAR_F1) && (c != '?'))
		{
			processchar(c, offset, offset_b, width, format,
			            scalefactor, progstate, redraw, end_program,
			            hide_hud, hud_hex, filesize);
		}

	}
	else if (*progstate == PROGSTATE_SET_WIDTH)
	{
		/* Add ASCII character */
		if (((c >= '0') && (c <= '9')) ||
		    (c == 'x') || (c == 'X') ||
		    ((c >= 'a') && (c <= 'f')) ||
		    ((c >= 'A') && (c <= 'F')) ||
		    (c == '+') || (c == '-'))
		{
			if ((inbufferpos + 0x0001) < INBUFSIZE)
			{
				inbuffer[inbufferpos] = c;
				++inbufferpos;
				inbuffer[inbufferpos] = '\0';
			}
		}
		/* Delete a character */
		else if (c == CHAR_BACKSPACE)
		{
			if (inbufferpos > 0x0000)
			{
				--inbufferpos;
			}
			inbuffer[inbufferpos] = '\0';
		}
		/* Cancel entering width */
		else if (c == CHAR_ESC)
		{
			inbufferpos = 0x0000;
			inbuffer[0] = '\0';
			*progstate = PROGSTATE_NORMAL;
		}
		/* Process entered width */
		else if (c == CHAR_NEWLINE)
		{
			/* Checks that anything was entered at all */
			if (inbufferpos != 0x0000)
			{
				operation = 0;
				temphex = 0;
				curpos = 0;
				
				/* Checks whether to set, add or subtract */
				if (inbuffer[0] == '-')
				{
					operation = 2;
					curpos++;
				}
				else if (inbuffer[0] == '+')
				{
					operation = 1;
					curpos++;
				}
			
				/* Checks whether the entered value is hexadecimal */
				if ((inbuffer[curpos] == 'x') || (inbuffer[curpos] == 'X'))
				{
					++curpos;
					temphex = 1;
				}
				else if ((inbuffer[(curpos + 1)] == 'x') ||
				         (inbuffer[(curpos + 1)] == 'X'))
				{
					curpos += 2;
					temphex = 1;
				}
				
				/* Reads the value */
				if (temphex == 1)
				{
					temp = (INTU32) strtol(&inbuffer[curpos], NULL, 16);
				}
				else
				{
					temp = (INTU32) strtol(&inbuffer[curpos], NULL, 10);
				}

				/* Performs the operation */
				if (operation == 0)
				{
					*width = temp;
				}
				else if (operation == 1)
				{
					*width += temp;
				}
				else if (operation == 2)
				{
					if (temp <= *width)
					{
						*width -= temp;
					}
					else
					{
						*width = 0x00000001;
					}
				}
			}
			inbufferpos = 0x0000;
			inbuffer[0] = '\0';
			*progstate = PROGSTATE_NORMAL;
		}
		(*redraw)++;
	}
	else if (*progstate == PROGSTATE_SET_OFFSET)
	{
		/* Add ASCII character */
		if (((c >= '0') && (c <= '9')) ||
		    (c == 'x') || (c == 'X') ||
		    ((c >= 'a') && (c <= 'f')) ||
		    ((c >= 'A') && (c <= 'F')) ||
		    (c == '+') || (c == '-'))
		{
			if ((inbufferpos + 0x0001) < INBUFSIZE)
			{
				inbuffer[inbufferpos] = c;
				++inbufferpos;
				inbuffer[inbufferpos] = '\0';
			}
		}
		/* Delete a character */
		else if (c == CHAR_BACKSPACE)
		{
			if (inbufferpos > 0x0000)
			{
				--inbufferpos;
			}
			inbuffer[inbufferpos] = '\0';
		}
		/* Cancel entering value */
		else if (c == CHAR_ESC)
		{
			inbufferpos = 0x0000;
			inbuffer[0] = '\0';
			*progstate = PROGSTATE_NORMAL;
		}
		/* Process entered value */
		else if (c == CHAR_NEWLINE)
		{
			/* Checks that anything was entered at all */
			if (inbufferpos != 0x0000)
			{
				operation = 0;
				temphex = 0;
				curpos = 0;
				
				/* Checks whether to set, add or subtract */
				if (inbuffer[0] == '-')
				{
					operation = 2;
					curpos++;
				}
				else if (inbuffer[0] == '+')
				{
					operation = 1;
					curpos++;
				}
			
				/* Checks whether the entered value is hexadecimal */
				if ((inbuffer[curpos] == 'x') || (inbuffer[curpos] == 'X'))
				{
					++curpos;
					temphex = 1;
				}
				else if ((inbuffer[(curpos + 1)] == 'x') ||
				         (inbuffer[(curpos + 1)] == 'X'))
				{
					curpos += 2;
					temphex = 1;
				}
				
				/* Reads the value */
				if (temphex == 1)
				{
					temp = (INTU32) strtol(&inbuffer[curpos], NULL, 16);
				}
				else
				{
					temp = (INTU32) strtol(&inbuffer[curpos], NULL, 10);
				}

				/* Performs the operation */
				if (operation == 0)
				{
					*offset = temp;
				}
				else if (operation == 1)
				{
					*offset += temp;
				}
				else if (operation == 2)
				{
					if (temp <= *offset)
					{
						*offset -= temp;
					}
					else
					{
						*offset = 0x00000001;
					}
				}
			}
			inbufferpos = 0x0000;
			inbuffer[0] = '\0';
			*progstate = PROGSTATE_NORMAL;
		}
		(*redraw)++;
	}

	/* Ensures the output variables make any sort of sense */
	if (*width < 0x00000001)
	{
		*width = 0x00000001;
	}
	maxWidth = (INTU32) pixelformats[*format];
	maxWidth &= 0x000000FF;
	maxWidth = ((filesize << 3U) / maxWidth);
	if (*width > maxWidth)
	{
		*width = maxWidth;
	}
	if (*offset > filesize)
	{
		*offset = filesize;
	}
}
