/*
 * Program to Display Binary Data as an Image (Linux Framebuffer Version)
 * Copyright (c) 2020 b122251 <hi@b122251.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of mosquitto nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* File Inclusions */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include "fixedint.h"
#include "ui.h"
#include "errors.h"
#include "hud.h"
#include "input.h"
#include "draw.h"

/* Macro Definitions */
#define MAX_FILESIZE 0x1FFFFFFF /* Maximum filesize (512 MB) */
#define LOOPTIME     25000      /* How long to wait before re-drawing */

/* Main Function of the Program */
int main(int argc, char *argv[])
{
	/* Variable Declarations */
	FILE *infile = NULL;     /* Pointer to the file to be read */
	INTU32 filesize;         /* Size of the input file */
	INTU8 *inmemfile = NULL; /* Space to read the file into memory */
	int fseekval;            /* Return value for fseek() */
	size_t freadval;         /* Return value for fread() */
	int fcloseval;           /* Return value for fclose() */
	int retval = ERROR_NONE; /* Return value for this function */

	/* Variables relating to the frame output */
	INTU8 *out_display = NULL; /* Where the frame will be stored */
	INTU32 out_width;          /* Width of the output frame (in pixels) */
	INTU32 out_height;         /* Height of the output frame (in pixels) */
	INTU16 out_format;         /* Pixel format of the output frame */

	/* Variables for drawing the frame */
	INTU32 in_offset;      /* Offset from where to start drawing (in bytes) */
	INTU8  in_offset_b;    /* Offset from where to start drawing (bits) */
	INTU32 in_width;       /* How many pixels wide to draw the file */
	INTU16 in_format;      /* Index of the pixel format to interpret the file */
	INTU32 in_scalefactor; /* How many times to double up each pixel */
	
	/* Variables relating to the flow of the program */
	pthread_t input_thread; /* Thread to process input in */
	struct input_thread_parameters inputparams; /* Input thread parameters */
	int progstate = 0;      /* State of the program */
	int inputthreadretval;  /* Return value for the input thread */
	INTU8 hide_hud;         /* Hide the HUD */
	INTU8 hud_hex;          /* Display the HUD values in hex */
	INTU8 term_inited;      /* Whether term_init() has already run */
	INTU8 end_program;      /* Set to 1 if the program must end */
	INTU8 redraw_screen;    /* Set to 1 if the screen needs to be redrawn */
	term_inited = (INTU8) 0x00; /* Whether the terminal is initialised */

	/* Initialises variables to 0 */
	out_width = 0x00000000;
	out_height = 0x00000000;
	out_format = 0x0000;
	in_offset = 0x00000000;
	in_offset_b = (INTU8) 0x00;
	in_width = 0x00000000;
	in_format = 0x0000;
	in_scalefactor = 0x00000000;
	inputthreadretval = 0;
	hide_hud = (INTU8) 0x00;
	hud_hex = (INTU8) 0x00;
	term_inited = (INTU8) 0x00;
	end_program = (INTU8) 0x00;
	redraw_screen = (INTU8) 0x00;
	
	/* Checks the number of command-line parameters */
	if (argc < 2)
	{
		retval = ERROR_PARAMETERS;
		goto end;
	}

	/* Opens up the file */
	infile = fopen(argv[1], "rb");
	if (infile == NULL)
	{
		retval = ERROR_OPEN_FILE_FAILED;
		goto end;
	}

	/* Reads filesize */
	fseekval = fseek(infile, 0l, SEEK_END);
	if (fseekval != 0)
	{
		retval = ERROR_READ_FILE_FAILED;
		goto end;
	}
	filesize = (INTU32) ftell(infile);

	/* Checks filesize */
	if (filesize > MAX_FILESIZE)
	{
		retval = ERROR_FILE_TOO_LARGE;
		goto end;
	}

	/* Allocates the space to load the file into memory */
	inmemfile = malloc((size_t) filesize);
	if (inmemfile == NULL)
	{
		retval = ERROR_MEMORY;
		goto end;
	}

	/* Reads the file into memory */
	fseekval = fseek(infile, 0l, SEEK_SET);
	if (fseekval != 0)
	{
		retval = ERROR_READ_FILE_FAILED;
		goto end;
	}
	freadval = fread(inmemfile, (size_t) 1, (size_t) filesize, infile);
	if (freadval != (size_t) filesize)
	{
		retval = ERROR_READ_FILE_FAILED;
		goto end;
	}

	/* Closes the input file */
	fcloseval = fclose(infile);
	if (fcloseval != 0)
	{
		retval = ERROR_CLOSE_FILE_FAILED;
		goto end;
	}
	infile = NULL;

	/* Makes the output display */
	retval = make_image_display(&out_display, &out_width,
	                            &out_height, &out_format, (INTU8) 0x01);
	if (retval != ERROR_NONE)
	{
		goto end;
	}

	/* Initialises the terminal to capture keyboard input */
	retval = terminal_init();
	if (retval != ERROR_NONE)
	{
		goto end;
	}
	term_inited = (INTU8) 0x01;
	
	/* Sets variables to defaults */
	in_offset = 0x00000000;
	in_offset_b = (INTU8) 0x07;
	in_width = 0x00000100;
	in_format = DEFAULT_PIXELFORMAT;
	in_scalefactor = 0x00000001;

	end_program = (INTU8) 0x00;
	redraw_screen = (INTU8) 0x01;
	
	/* Starts input processing thread */
	inputparams.offset = &in_offset;
	inputparams.offset_b = &in_offset_b;
	inputparams.width = &in_width;
	inputparams.format = &in_format;
	inputparams.scalefactor = &in_scalefactor;
	inputparams.progstate = &progstate;
	inputparams.redraw = &redraw_screen;
	inputparams.end_program = &end_program;
	inputparams.hide_hud = &hide_hud;
	inputparams.hud_hex = &hud_hex;
	inputparams.retval = &inputthreadretval;
	inputparams.filesize = filesize;
	(void) pthread_create(&input_thread, NULL,
	                      (void *) inputThread, &inputparams);

	/* Loop that draws the screen */
	while (end_program == (INTU8) 0x00)
	{
		/* Ends the program if the other thread has crashed */
		if (inputthreadretval != ERROR_NONE)
		{
			break;
		}

		/* Re-draws the screen if needed */
		if (redraw_screen > (INTU8) 0x00)
		{
			retval = drawscreen(inmemfile, filesize, in_offset, in_offset_b,
			                    in_width, pixelformats[in_format], out_display,
			                    out_width, out_height, out_format,
			                    in_scalefactor);
			if (retval != ERROR_NONE)
			{
				end_program = (INTU8) 0x01;
				goto end;
			}

			/* Draws the HUD over the screen */
			retval = displayHUD(out_display, out_width, out_height, out_format,
			                    in_offset, in_offset_b, in_width, in_format,
			                    in_scalefactor, progstate, hide_hud, hud_hex);
			if (retval != ERROR_NONE)
			{
				end_program = (INTU8) 0x01;
				goto end;
			}

			/* Remembers that the screen has been drawn */
			redraw_screen--;
		}

		/* Pushes the screen to the framebuffer */
		retval = write_display_to_output(out_display, out_width, out_height,
		                                 out_format, (INTU8) 0x01);
		if (retval != ERROR_NONE)
		{
			end_program = (INTU8) 0x01;
			goto end;
		}

		/* Sleeps away the rest of the looptime */
		(void) usleep(LOOPTIME);
	}

end:/* Clean up before ending the program */
	/* De-allocates the display */
	(void) unmake_image_display(&out_display, (INTU8) 0x01);

	/* Resets the original terminal settings */
	if (term_inited == (INTU8) 0x01)
	{
		(void) terminal_uninit();
	}

	/* Closes the input file if it's still open */
	if (infile != NULL)
	{
		fcloseval = fclose(infile);
		if (fcloseval != 0)
		{
			retval = ERROR_CLOSE_FILE_FAILED;
		}
	}

	/* Frees the in-memory copy of the input file */
	if (inmemfile != NULL)
	{
		free(inmemfile);
	}

	/* Prints an error message if needed */
	if (retval != ERROR_NONE)
	{
		printError(retval, argv[1]);
	}

	/* Ends the program */
	return retval;
}
