/*
 * Library to Overlay Text on the Framebuffer Image
 * Copyright (c) 2020 b122251 <hi@b122251.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of mosquitto nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* File Inclusions */
#include <stdio.h>
#include <string.h>
#include "version.h"
#include "fixedint.h"
#include "errors.h"
#include "input.h"
#include "draw.h"
#include "hud.h"

/*
 * Definition of a 1-bit bitmapped monospace font.
 *
 * The font consists of 96 characters (all 95 printable ASCII-characters, and a placeholder
 * that can be used for any invalid character (here filled with a dither pattern).
 * Every character is 8 pixels wide, and 16 pixels high. A character consists of a 16-byte
 * array. Every byte represents a row of pixels, in which if a bit is set, it is part of
 * the character, and if the bit is not set, it is not part of the character.
 */
static INTU8 font[96][16] = {
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, /* Space */
	{0x00, 0x00, 0x18, 0x3C, 0x3C, 0x3C, 0x18, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00}, /* ! */
	{0x00, 0x66, 0x66, 0x66, 0x24, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, /* " */
	{0x00, 0x00, 0x00, 0x6C, 0x6C, 0xFE, 0x6C, 0x6C, 0x6C, 0xFE, 0x6C, 0x6C, 0x00, 0x00, 0x00, 0x00}, /* # */
	{0x18, 0x18, 0x7C, 0xC6, 0xC2, 0xC0, 0x7C, 0x06, 0x06, 0x86, 0xC6, 0x7C, 0x18, 0x18, 0x00, 0x00}, /* $ */
	{0x00, 0x00, 0x00, 0x00, 0xC2, 0xC6, 0x0C, 0x18, 0x30, 0x60, 0xC6, 0x86, 0x00, 0x00, 0x00, 0x00}, /* % */
	{0x00, 0x00, 0x38, 0x6C, 0x6C, 0x38, 0x76, 0xDC, 0xCC, 0xCC, 0xCC, 0x76, 0x00, 0x00, 0x00, 0x00}, /* & */
	{0x00, 0x30, 0x30, 0x30, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, /* ' */
	{0x00, 0x00, 0x0C, 0x18, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x18, 0x0C, 0x00, 0x00, 0x00, 0x00}, /* ( */
	{0x00, 0x00, 0x30, 0x18, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x18, 0x30, 0x00, 0x00, 0x00, 0x00}, /* ) */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x66, 0x3C, 0xFF, 0x3C, 0x66, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, /* * */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x7E, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, /* + */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x18, 0x30, 0x00, 0x00, 0x00}, /* , */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFE, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, /* - */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00}, /* . */
	{0x00, 0x00, 0x00, 0x00, 0x02, 0x06, 0x0C, 0x18, 0x30, 0x60, 0xC0, 0x80, 0x00, 0x00, 0x00, 0x00}, /* / */
	{0x00, 0x00, 0x38, 0x6C, 0xC6, 0xC6, 0xD6, 0xD6, 0xC6, 0xC6, 0x6C, 0x38, 0x00, 0x00, 0x00, 0x00}, /* 0 */
	{0x00, 0x00, 0x18, 0x38, 0x78, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x7E, 0x00, 0x00, 0x00, 0x00}, /* 1 */
	{0x00, 0x00, 0x7C, 0xC6, 0x06, 0x0C, 0x18, 0x30, 0x60, 0xC0, 0xC6, 0xFE, 0x00, 0x00, 0x00, 0x00}, /* 2 */
	{0x00, 0x00, 0x7C, 0xC6, 0x06, 0x06, 0x3C, 0x06, 0x06, 0x06, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00}, /* 3 */
	{0x00, 0x00, 0x0C, 0x1C, 0x3C, 0x6C, 0xCC, 0xFE, 0x0C, 0x0C, 0x0C, 0x1E, 0x00, 0x00, 0x00, 0x00}, /* 4 */
	{0x00, 0x00, 0xFE, 0xC0, 0xC0, 0xC0, 0xFC, 0x06, 0x06, 0x06, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00}, /* 5 */
	{0x00, 0x00, 0x38, 0x60, 0xC0, 0xC0, 0xFC, 0xC6, 0xC6, 0xC6, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00}, /* 6 */
	{0x00, 0x00, 0xFE, 0xC6, 0x06, 0x06, 0x0C, 0x18, 0x30, 0x30, 0x30, 0x30, 0x00, 0x00, 0x00, 0x00}, /* 7 */
	{0x00, 0x00, 0x7C, 0xC6, 0xC6, 0xC6, 0x7C, 0xC6, 0xC6, 0xC6, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00}, /* 8 */
	{0x00, 0x00, 0x7C, 0xC6, 0xC6, 0xC6, 0x7E, 0x06, 0x06, 0x06, 0x0C, 0x78, 0x00, 0x00, 0x00, 0x00}, /* 9 */
	{0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00}, /* : */
	{0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x18, 0x18, 0x30, 0x00, 0x00, 0x00, 0x00}, /* ; */
	{0x00, 0x00, 0x00, 0x06, 0x0C, 0x18, 0x30, 0x60, 0x30, 0x18, 0x0C, 0x06, 0x00, 0x00, 0x00, 0x00}, /* < */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x7E, 0x00, 0x00, 0x7E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, /* = */
	{0x00, 0x00, 0x00, 0x60, 0x30, 0x18, 0x0C, 0x06, 0x0C, 0x18, 0x30, 0x60, 0x00, 0x00, 0x00, 0x00}, /* > */
	{0x00, 0x00, 0x7C, 0xC6, 0xC6, 0x0C, 0x18, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00}, /* ? */
	{0x00, 0x00, 0x00, 0x7C, 0xC6, 0xC6, 0xDE, 0xDE, 0xDE, 0xDC, 0xC0, 0x7C, 0x00, 0x00, 0x00, 0x00}, /* @ */
	{0x00, 0x00, 0x10, 0x38, 0x6C, 0xC6, 0xC6, 0xFE, 0xC6, 0xC6, 0xC6, 0xC6, 0x00, 0x00, 0x00, 0x00}, /* A */
	{0x00, 0x00, 0xFC, 0x66, 0x66, 0x66, 0x7C, 0x66, 0x66, 0x66, 0x66, 0xFC, 0x00, 0x00, 0x00, 0x00}, /* B */
	{0x00, 0x00, 0x3C, 0x66, 0xC2, 0xC0, 0xC0, 0xC0, 0xC0, 0xC2, 0x66, 0x3C, 0x00, 0x00, 0x00, 0x00}, /* C */
	{0x00, 0x00, 0xF8, 0x6C, 0x66, 0x66, 0x66, 0x66, 0x66, 0x66, 0x6C, 0xF8, 0x00, 0x00, 0x00, 0x00}, /* D */
	{0x00, 0x00, 0xFE, 0x66, 0x62, 0x68, 0x78, 0x68, 0x60, 0x62, 0x66, 0xFE, 0x00, 0x00, 0x00, 0x00}, /* E */
	{0x00, 0x00, 0xFE, 0x66, 0x62, 0x68, 0x78, 0x68, 0x60, 0x60, 0x60, 0xF0, 0x00, 0x00, 0x00, 0x00}, /* F */
	{0x00, 0x00, 0x3C, 0x66, 0xC2, 0xC0, 0xC0, 0xDE, 0xC6, 0xC6, 0x66, 0x3A, 0x00, 0x00, 0x00, 0x00}, /* G */
	{0x00, 0x00, 0xC6, 0xC6, 0xC6, 0xC6, 0xFE, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x00, 0x00, 0x00, 0x00}, /* H */
	{0x00, 0x00, 0x3C, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3C, 0x00, 0x00, 0x00, 0x00}, /* I */
	{0x00, 0x00, 0x1E, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0xCC, 0xCC, 0xCC, 0x78, 0x00, 0x00, 0x00, 0x00}, /* J */
	{0x00, 0x00, 0xE6, 0x66, 0x66, 0x6C, 0x78, 0x78, 0x6C, 0x66, 0x66, 0xE6, 0x00, 0x00, 0x00, 0x00}, /* K */
	{0x00, 0x00, 0xF0, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x62, 0x66, 0xFE, 0x00, 0x00, 0x00, 0x00}, /* L */
	{0x00, 0x00, 0xC6, 0xEE, 0xFE, 0xFE, 0xD6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x00, 0x00, 0x00, 0x00}, /* M */
	{0x00, 0x00, 0xC6, 0xE6, 0xF6, 0xFE, 0xDE, 0xCE, 0xC6, 0xC6, 0xC6, 0xC6, 0x00, 0x00, 0x00, 0x00}, /* N */
	{0x00, 0x00, 0x7C, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00}, /* O */
	{0x00, 0x00, 0xFC, 0x66, 0x66, 0x66, 0x7C, 0x60, 0x60, 0x60, 0x60, 0xF0, 0x00, 0x00, 0x00, 0x00}, /* P */
	{0x00, 0x00, 0x7C, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xD6, 0xDE, 0x7C, 0x0C, 0x0E, 0x00, 0x00}, /* Q */
	{0x00, 0x00, 0xFC, 0x66, 0x66, 0x66, 0x7C, 0x6C, 0x66, 0x66, 0x66, 0xE6, 0x00, 0x00, 0x00, 0x00}, /* R */
	{0x00, 0x00, 0x7C, 0xC6, 0xC6, 0x60, 0x38, 0x0C, 0x06, 0xC6, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00}, /* S */
	{0x00, 0x00, 0x7E, 0x7E, 0x5A, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3C, 0x00, 0x00, 0x00, 0x00}, /* T */
	{0x00, 0x00, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00}, /* U */
	{0x00, 0x00, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x6C, 0x38, 0x10, 0x00, 0x00, 0x00, 0x00}, /* V */
	{0x00, 0x00, 0xC6, 0xC6, 0xC6, 0xC6, 0xD6, 0xD6, 0xD6, 0xFE, 0xEE, 0x6C, 0x00, 0x00, 0x00, 0x00}, /* W */
	{0x00, 0x00, 0xC6, 0xC6, 0x6C, 0x7C, 0x38, 0x38, 0x7C, 0x6C, 0xC6, 0xC6, 0x00, 0x00, 0x00, 0x00}, /* X */
	{0x00, 0x00, 0x66, 0x66, 0x66, 0x66, 0x3C, 0x18, 0x18, 0x18, 0x18, 0x3C, 0x00, 0x00, 0x00, 0x00}, /* Y */
	{0x00, 0x00, 0xFE, 0xC6, 0x86, 0x0C, 0x18, 0x30, 0x60, 0xC2, 0xC6, 0xFE, 0x00, 0x00, 0x00, 0x00}, /* Z */
	{0x00, 0x00, 0x3C, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x3C, 0x00, 0x00, 0x00, 0x00}, /* [ */
	{0x00, 0x00, 0x00, 0x80, 0xC0, 0xE0, 0x70, 0x38, 0x1C, 0x0E, 0x06, 0x02, 0x00, 0x00, 0x00, 0x00}, /* \ */
	{0x00, 0x00, 0x3C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x3C, 0x00, 0x00, 0x00, 0x00}, /* ] */
	{0x10, 0x38, 0x6C, 0xC6, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, /* ^ */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00}, /* _ */
	{0x30, 0x30, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, /* ` */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x78, 0x0C, 0x7C, 0xCC, 0xCC, 0xCC, 0x76, 0x00, 0x00, 0x00, 0x00}, /* a */
	{0x00, 0x00, 0xE0, 0x60, 0x60, 0x78, 0x6C, 0x66, 0x66, 0x66, 0x66, 0x7C, 0x00, 0x00, 0x00, 0x00}, /* b */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x7C, 0xC6, 0xC0, 0xC0, 0xC0, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00}, /* c */
	{0x00, 0x00, 0x1C, 0x0C, 0x0C, 0x3C, 0x6C, 0xCC, 0xCC, 0xCC, 0xCC, 0x76, 0x00, 0x00, 0x00, 0x00}, /* d */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x7C, 0xC6, 0xFE, 0xC0, 0xC0, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00}, /* e */
	{0x00, 0x00, 0x38, 0x6C, 0x64, 0x60, 0xF0, 0x60, 0x60, 0x60, 0x60, 0xF0, 0x00, 0x00, 0x00, 0x00}, /* f */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x76, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0x7C, 0x0C, 0xCC, 0x78, 0x00}, /* g */
	{0x00, 0x00, 0xE0, 0x60, 0x60, 0x6C, 0x76, 0x66, 0x66, 0x66, 0x66, 0xE6, 0x00, 0x00, 0x00, 0x00}, /* h */
	{0x00, 0x00, 0x18, 0x18, 0x00, 0x38, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3C, 0x00, 0x00, 0x00, 0x00}, /* i */
	{0x00, 0x00, 0x06, 0x06, 0x00, 0x0E, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x66, 0x66, 0x3C, 0x00}, /* j */
	{0x00, 0x00, 0xE0, 0x60, 0x60, 0x66, 0x6C, 0x78, 0x78, 0x6C, 0x66, 0xE6, 0x00, 0x00, 0x00, 0x00}, /* k */
	{0x00, 0x00, 0x38, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3C, 0x00, 0x00, 0x00, 0x00}, /* l */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0xEC, 0xFE, 0xD6, 0xD6, 0xD6, 0xD6, 0xC6, 0x00, 0x00, 0x00, 0x00}, /* m */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0xDC, 0x66, 0x66, 0x66, 0x66, 0x66, 0x66, 0x00, 0x00, 0x00, 0x00}, /* n */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x7C, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00}, /* o */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0xDC, 0x66, 0x66, 0x66, 0x66, 0x66, 0x7C, 0x60, 0x60, 0xF0, 0x00}, /* p */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x76, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0x7C, 0x0C, 0x0C, 0x1E, 0x00}, /* q */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0xDC, 0x76, 0x66, 0x60, 0x60, 0x60, 0xF0, 0x00, 0x00, 0x00, 0x00}, /* r */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x7C, 0xC6, 0x60, 0x38, 0x0C, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00}, /* s */
	{0x00, 0x00, 0x10, 0x30, 0x30, 0xFC, 0x30, 0x30, 0x30, 0x30, 0x36, 0x1C, 0x00, 0x00, 0x00, 0x00}, /* t */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0x76, 0x00, 0x00, 0x00, 0x00}, /* u */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x66, 0x66, 0x66, 0x66, 0x66, 0x3C, 0x18, 0x00, 0x00, 0x00, 0x00}, /* v */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0xC6, 0xC6, 0xD6, 0xD6, 0xD6, 0xFE, 0x6C, 0x00, 0x00, 0x00, 0x00}, /* w */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0xC6, 0x6C, 0x38, 0x38, 0x38, 0x6C, 0xC6, 0x00, 0x00, 0x00, 0x00}, /* x */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x7E, 0x06, 0x0C, 0xF8, 0x00}, /* y */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0xFE, 0xCC, 0x18, 0x30, 0x60, 0xC6, 0xFE, 0x00, 0x00, 0x00, 0x00}, /* z */
	{0x00, 0x00, 0x0E, 0x18, 0x18, 0x18, 0x70, 0x18, 0x18, 0x18, 0x18, 0x0E, 0x00, 0x00, 0x00, 0x00}, /* { */
	{0x00, 0x00, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00}, /* | */
	{0x00, 0x00, 0x70, 0x18, 0x18, 0x18, 0x0E, 0x18, 0x18, 0x18, 0x18, 0x70, 0x00, 0x00, 0x00, 0x00}, /* } */
	{0x00, 0x00, 0x76, 0xDC, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, /* ~ */
	{0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55}  /* Invalid */
};

/* Macro's for defining foreground and background pixels */
#define fgPixel_0020(a) \
{                       \
	a = 0x00FFFFFF;     \
}
#define bgPixel_0020(a) \
{                       \
	a &= 0x00F8F8F8;    \
	a >>= 3U;           \
}

/* The input text buffer */
extern char inbuffer[INBUFSIZE];
extern INTU16 inbufferpos;

static void printchar_0020(INTU32 *display, INTU32 width, char character)
{
	/* Variable Declarations */
	int lineIndex;         /* Current row of pixels in the character */
	INTU8 *fontchar; /* Pointer to the bitmap of the character */
	
	/* Sets fontchar */
	fontchar = &font[(int) (character - ' ')][0];

	/* Draws the character */
	for (lineIndex = 0; lineIndex < 16; ++lineIndex)
	{
		if ((*fontchar & (INTU8) 0x80) == (INTU8) 0x00)
		{
			bgPixel_0020(display[0]);
		}
		else
		{
			fgPixel_0020(display[0]);
		}
		if ((*fontchar & (INTU8) 0x40) == (INTU8) 0x00)
		{
			bgPixel_0020(display[1]);
		}
		else
		{
			fgPixel_0020(display[1]);
		}
		if ((*fontchar & (INTU8) 0x20) == (INTU8) 0x00)
		{
			bgPixel_0020(display[2]);
		}
		else
		{
			fgPixel_0020(display[2]);
		}
		if ((*fontchar & (INTU8) 0x10) == (INTU8) 0x00)
		{
			bgPixel_0020(display[3]);
		}
		else
		{
			fgPixel_0020(display[3]);
		}
		if ((*fontchar & (INTU8) 0x08) == (INTU8) 0x00)
		{
			bgPixel_0020(display[4]);
		}
		else
		{
			fgPixel_0020(display[4]);
		}
		if ((*fontchar & (INTU8) 0x04) == (INTU8) 0x00)
		{
			bgPixel_0020(display[5]);
		}
		else
		{
			fgPixel_0020(display[5]);
		}
		if ((*fontchar & (INTU8) 0x02) == (INTU8) 0x00)
		{
			bgPixel_0020(display[6]);
		}
		else
		{
			fgPixel_0020(display[6]);
		}
		if ((*fontchar & (INTU8) 0x01) == (INTU8) 0x00)
		{
			bgPixel_0020(display[7]);
		}
		else
		{
			fgPixel_0020(display[7]);
		}
		display = &(display[width]);
		++fontchar;
	}
}

int string_to_display(INTU8 *display, char *string, INTU16 stringlen,
                      INTU32 width, INTU32 height, INTU16 format, INTU32 x,
                      INTU32 y, INTU16 alignment)
{
	/* Variable Declarations */
	INTU16 curchar;      /* Current character in processing */
	INTU32 lineWidth;    /* With of the current line (in pixels) */
	INTU32 stringWidth;  /* Width of the string (in pixels) */
	INTU32 stringHeight; /* Height of the string (in pixels) */
	INTU32 startpos_x;   /* X-pixel where to start writing the text */
	INTU32 startpos_y;   /* Y-pixel where to start writing the text */
	INTU32 charhBytes;   /* Number of bytes to 16 pixels below */
	INTU32 charwBytes;   /* Size of 8 pixels (in bytes) */
	INTU32 lineOffset;   /* Offset where the current line started */
	INTU32 curpos;       /* Current position in the display */
	/* Pointer to the function that prints a character */
	void (*printchar)(void *, INTU32, char) = NULL;

	/* Determines printchar */
	switch (format)
	{
		case PIXELFORMAT_16BIT_XRGB1555:
			break;
		case PIXELFORMAT_24BIT_RGB888:
			break;
		case PIXELFORMAT_32BIT_XRGB8888:
			printchar = (void *) &printchar_0020;
			break;
	}
	if (printchar == NULL)
	{
		printf(PROGNAME ": No printchar() function was found\n");
		return ERROR_OSDEP;
	}

	/* Determines the size of the string (in pixels) */
	stringWidth = 0x00000000;
	stringHeight = 0x00000010;
	lineWidth = 0x00000000;
	for (curchar = 0x0000; curchar < stringlen; ++curchar)
	{
		if ((string[curchar] >= ' ') && (string[curchar] <= '~'))
		{
			lineWidth += 0x00000008;
		}
		else if (string[curchar] == '\n')
		{
			stringHeight += 0x00000010;
			if (lineWidth > stringWidth)
			{
				stringWidth = lineWidth;
			}
			lineWidth = 0x00000000;
		}
	}
	if (lineWidth > stringWidth)
	{
		stringWidth = lineWidth;
	}
	
	/* Checks that the text is going to fit */
	if (((stringWidth + x) > width) ||
	    ((stringHeight + y) > height))
	{
		printf(PROGNAME ": Text was too large to print\n");
		return ERROR_OSDEP;
	}

	/* Caluclates where to start drawing */
	startpos_x = 0x00000000;
	startpos_y = 0x00000000;
	if ((alignment & HUD_ALIGN_X) == HUD_ALIGN_XL)
	{
		startpos_x = x;
	}
	else if ((alignment & HUD_ALIGN_X) == HUD_ALIGN_XM)
	{
		startpos_x = (width >> 1U);
		startpos_x -= (stringWidth >> 1U);
	}
	else if ((alignment & HUD_ALIGN_X) == HUD_ALIGN_XR)
	{
		startpos_x = (width - x);
		startpos_x -= 0x00000008;
	}
	if ((alignment & HUD_ALIGN_Y) == HUD_ALIGN_YT)
	{
		startpos_y = y;
	}
	else if ((alignment & HUD_ALIGN_Y) == HUD_ALIGN_YM)
	{
		startpos_y = (height >> 1U);
		startpos_y -= (stringHeight >> 1U);
	}
	else if ((alignment & HUD_ALIGN_Y) == HUD_ALIGN_YB)
	{
		startpos_y = (height - y);
		startpos_y -= 0x00000010;
	}

	/* Determines charwBytes, charhBytes and lineOffset */
	charwBytes = (INTU32) (format & 0x00FF);
	charhBytes = (charwBytes * width);
	charhBytes <<= 1U;
	lineOffset = (startpos_y * width);
	lineOffset += startpos_x;
	lineOffset *= (INTU32) ((format & 0x00FF) >> 3U);
	curpos = lineOffset;
	
	/* Prints forwards or backwards based on justification */
	if ((alignment & HUD_JUSTIFIC) == HUD_JUSTIF_L)
	{
		for (curchar = 0x0000; curchar < stringlen; ++curchar)
		{
			/* Prints the character if it is printable */
			if ((string[curchar] >= ' ') && (string[curchar] <= '~'))
			{
				(*printchar)(&display[curpos], width ,string[curchar]);
				curpos += charwBytes;
			}
			/* Goes to the next line if this is a newline */
			else if (string[curchar] == '\n')
			{
				if ((alignment & HUD_ALIGN_Y) == (HUD_ALIGN_YB))
				{
					lineOffset -= charhBytes;
				}
				else
				{
					lineOffset += charhBytes;
				}
				curpos = lineOffset;
			}
		}
	}
	else if ((alignment & HUD_JUSTIFIC) == HUD_JUSTIF_R)
	{
		do
		{
			--curchar;
			/* Prints the character if it is printable */
			if ((string[curchar] >= ' ') && (string[curchar] <= '~'))
			{
				(*printchar)(&display[curpos], width ,string[curchar]);
				curpos -= charwBytes;
			}
			/* Goes to the next line if this is a newline */
			else if (string[curchar] == '\n')
			{
				if ((alignment & HUD_ALIGN_Y) == (HUD_ALIGN_YB))
				{
					lineOffset -= charhBytes;
				}
				else
				{
					lineOffset += charhBytes;
				}
				curpos = lineOffset;
			}
		}
		while (curchar > 0x0000);
	}

	return ERROR_NONE;
}

/*
 * Function that displays the HUD onto the screen.
 * Parameters:
 * * out_display = Pointer to the display
 * * out_width = Width of the display (in pixels)
 * * out_height = Height of the display (in pixels)
 * * out_pixelformat = Pixel encoding format of the display
 * * offset = Current offset of the data
 * * offset_b = Current bit in the offset byte (if applicable)
 * * width = Current width
 * * format = Current format index
 * * scalefactor = Current zoom level
 * * progstate = State of the program
 * * hide_hud = Doesn't display the hud if not 0
 * * hud_hex = Whether to print values as hexadecimal
 */
int displayHUD(INTU8 *out_display, INTU32 out_width, INTU32 out_height,
               INTU16 out_pixelformat, INTU32 offset, INTU8 offset_b,
               INTU32 width, INTU16 format, INTU32 scalefactor, int progstate,
               INTU8 hide_hud, INTU8 hud_hex)
{
	/* Variable Declaratons */
	int retval = ERROR_NONE; /* Return value for this function */
	char outtext[128];
	char help_message[] =
		"                                                          \n"
		" Controls:                                                \n"
		"                                                          \n"
		" Up      k          -  Scroll one row of pixels up        \n"
		" Down    j          -  Scroll one row of pixels down      \n"
		" Left               -  Scroll one pixel to the left       \n"
		" Right              -  Scroll one pixel to the right      \n"
		" h                  -  Move the offset one byte forward   \n"
		" l                  -  Move the offset one byte backward  \n"
		" PgUp    K          -  Scroll sixteen rows of pixels up   \n"
		" PgDown  J          -  Scroll sixteen rows of pixels down \n"
		" o       O          -  Set offset                         \n"
		" H                  -  Subtract one pixel from the width  \n"
		" L                  -  Add one pixel to the width         \n"
		" w       W          -  Set width                          \n"
		" +                  -  Zoom in                            \n"
		" -                  -  Zoom out                           \n"
		" 0                  -  Reset zoom                         \n"
		" p                  -  Next pixel encoding format         \n"
		" P                  -  Previous pixel encoding format     \n"
		" x       X          -  Toggle decimal/hexademical display \n"
		" v       V          -  Toggle display                     \n"
		" F1      ?          -  Display this screen                \n"
		" Esc     Q       q  -  Quit the program                   \n"
		"                                                          ";

	/* Prints the special case parts of the hud */
	switch (progstate)
	{
		case PROGSTATE_HELPSCREEN:
			retval = string_to_display(out_display, help_message,
			                           (INTU16) strlen(help_message), out_width,
			                           out_height, out_pixelformat, 0x00000000,
			                           0x00000000, HUD_ALIGN_XM | HUD_ALIGN_YM);
			break;
		case PROGSTATE_SET_WIDTH:
			(void) sprintf(outtext, "Width: %s", inbuffer);
			retval = string_to_display(out_display, outtext, (INTU16)
			                           strlen(outtext), out_width, out_height,
			                           out_pixelformat, 0x00000000, 0x00000000,
									   HUD_ALIGN_XL | HUD_ALIGN_YB);
			break;
		case PROGSTATE_SET_OFFSET:
			(void) sprintf(outtext, "Offset: %s", inbuffer);
			retval = string_to_display(out_display, outtext, (INTU16)
			                           strlen(outtext), out_width, out_height,
			                           out_pixelformat, 0x00000000, 0x00000000,
									   HUD_ALIGN_XL | HUD_ALIGN_YB);
			break;
		case PROGSTATE_SET_SAVEPATH:
			(void) sprintf(outtext, "Save Screenshot As: %s", inbuffer);
			retval = string_to_display(out_display, outtext, (INTU16)
			                           strlen(outtext), out_width, out_height,
			                           out_pixelformat, 0x00000000, 0x00000000,
									   HUD_ALIGN_XL | HUD_ALIGN_YB);
			break;
	}

	/* Prints the usual HUD if needed */
	if ((retval == ERROR_NONE) && (hide_hud == (INTU8) 0x00))
	{
		if (hud_hex == (INTU8) 0x00)
		{
			/* Prints the numbers in decimal */
			(void) sprintf(outtext, "Format: %s\n"
			                        "Offset: " PRINTU32 "\n"
			                        "Width: " PRINTU32,
			               pixelformatnames[format], offset, width);
		}
		else
		{
			/* Prints the numbers in hexadecimal */
			(void) sprintf(outtext, "Format: %s\n"
			                        "Offset: 0x%06X\n"
			                        "Width: 0x%X",
			               pixelformatnames[format], offset, width);
		}
		retval = string_to_display(out_display, outtext, (INTU16)
		                           strlen(outtext), out_width, out_height,
		                           out_pixelformat, 0x00000000, 0x00000000,
								   HUD_ALIGN_XR | HUD_ALIGN_YT | HUD_JUSTIF_R);
		if (scalefactor != 0x00000001)
		{
			(void) sprintf(outtext, "Zoom: " PRINTU32 "X", scalefactor);
			retval = string_to_display(out_display, outtext, (INTU16)
			                           strlen(outtext), out_width, out_height,
			                           out_pixelformat, 0x00000000, 0x00000030,
									   HUD_ALIGN_XR | HUD_ALIGN_YT |
			                           HUD_JUSTIF_R);
		}
	}

	return retval;
}
