/*
 * Library for Interaction with the Linux Framebuffer
 * Copyright (c) 2020 b122251 <hi@b122251.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of mosquitto nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* File Inclusions */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <linux/fb.h>
#include <sys/ioctl.h>
#include "fixedint.h"
#include "errors.h"
#include "version.h"
#include "draw.h"

/* Internal Macro's */
#define FBDEV "/dev/fb0"

/* Internal Function Declarations */
static int getResolution(INTU32 *width, INTU32 *height, INTU16 *format);

/* Internal static variables */
static FILE *framebuffer;

/* Internal Functions */

/*
 * Function that gets the screen resolution on a Linux-machine.
 * Parameters:
 * * width = Where to store the width
 * * height = Where to store the height
 * * format = Where to store the pixel format
 * Return values:
 * * 0 = Resolution was found
 * * !0 = Resolution couldn't be determined
 */
static int getResolution(INTU32 *width, INTU32 *height, INTU16 *format)
{
	/* Variable Declarations */
    struct fb_var_screeninfo vinfo; /* Struct for reading resolution */
	int fd = -1;    /* File Descriptor for the frame buffer */
	int retval = 0; /* Return value for this function */

	/* Open the file for reading and writing */
    fd = open(FBDEV, O_RDWR);
    if (fd == -1)
	{
		retval = 1;
		goto end;
    }

	/* Gets screen info */
	retval = ioctl(fd, FBIOGET_VSCREENINFO, &vinfo);
	if (retval != 0)
	{
		retval = 2;
		goto end;
	}
	
	/* Gets the resolution from the read data */
	*width = (INTU32) vinfo.xres;
	*height = (INTU32) vinfo.yres;
	
	/* Determines the format */
	if ((vinfo.red.offset == (__u32) 16) &&
        (vinfo.red.length == (__u32) 8) &&
	    (vinfo.red.msb_right == (__u32) 0) &&
	    (vinfo.green.offset == (__u32) 8) &&
        (vinfo.green.length == (__u32) 8) &&
	    (vinfo.green.msb_right == (__u32) 0) &&
	    (vinfo.blue.offset == (__u32) 0) &&
        (vinfo.blue.length == (__u32) 8) &&
	    (vinfo.blue.msb_right == (__u32) 0) &&
		(vinfo.bits_per_pixel == (__u32) 32))
	{
		*format = PIXELFORMAT_32BIT_XRGB8888;
	}
	else
	{
		*format = 0x0000;
	}

end:/* Closes the framebuffer's file descriptor */
    close(fd);
	return retval;
}

/* External Function */
/*
 * Function that allocates the display area for the image.
 * Parameters:
 * * display     = Where to store the pointer to the display
 * * width       = Where to store the width of the display
 * * height      = Where to store the height of the display
 * * format      = Pixel format 
 * * printErrors = Whether to print errors (0 = No, 1 = Yes)
 */
int make_image_display(INTU8 **display, INTU32 *width,
                       INTU32 *height, INTU16 *format, INTU8 printErrors)
{
	/* Variable Declarations */
	int freeint1; /* Integer for general use */
	INTU32 freeuint321;

	/* Determines the resolution */
	freeint1 = getResolution(width, height, format);
	if (freeint1 != 0)
	{
		if (printErrors == (INTU8) 0x01)
		{
			if (freeint1 == 1)
			{
				printf(PROGNAME ": " FBDEV " couldn't be opened\n");
			}
			else if (freeint1 == 2)
			{
				printf(PROGNAME ": ioctl failed on " FBDEV "\n");
			}
		}
		return ERROR_OSDEP;
	}

	/* Allocates the display */
	freeuint321 = (INTU32) (*format & 0x00FF);
	freeuint321 >>= 3U;
	freeuint321 *= *width;
	freeuint321 *= *height;
	*display = calloc((size_t) freeuint321, (size_t) 1);
	if (*display == NULL)
	{
		return ERROR_MEMORY;
	}

	/* Opens up the framebuffer for reading and writing */
	framebuffer = NULL;
	framebuffer = fopen(FBDEV, "r+b");
	if (framebuffer == NULL)
	{
		if (printErrors == (INTU8) 0x01)
		{
			printf(PROGNAME ": " FBDEV " could not be opened\n");
		}
		return ERROR_OSDEP;
	}

	return ERROR_NONE;
}

/*
 * Function that destroys the image display
 * Parameters:
 * * display = Where the display is stored
 * * printErrors = Whether to print errors (0 = No, 1 = Yes)
 */
int unmake_image_display(INTU8 **display, INTU8 printErrors)
{
	/* Variable Declarations */
	int fd;        /* File descriptor for the framebuffer device */
	int fcloseval; /* Return value for fclose */
    struct fb_var_screeninfo vinfo; /* Struct for reading resolution */

	/* Frees the allocated display */
	if (*display != NULL)
	{
		free(*display);
		*display = NULL;
	}

	/* Closes the framebuffer */
	if (framebuffer != NULL)
	{
		fcloseval = fclose(framebuffer);
		if (fcloseval != 0)
		{
			if (printErrors == (INTU8) 0x01)
			{
				printf(PROGNAME ": " FBDEV " could not be closed\n");
			}
			return ERROR_OSDEP;
		}
		framebuffer = NULL;

		/* Changes the framebuffer's settings to ensure it gets redrawn */
	    fd = open(FBDEV, O_RDWR);
	    if (fd != -1)
		{
			/* Flips the order for transparency */
			(void) ioctl(fd, FBIOGET_VSCREENINFO, &vinfo);
			vinfo.transp.msb_right ^= 0x00000001;
			(void) ioctl(fd, FBIOPUT_VSCREENINFO, &vinfo);
			/* Flips it back */
			vinfo.transp.msb_right ^= 0x00000001;
			(void) ioctl(fd, FBIOPUT_VSCREENINFO, &vinfo);
			/* Closes the framebuffer */
			(void) close(fd);
		}
	}

	return ERROR_NONE;
}

/*
 * Function that writes the display to the actual display hardware.
 * Parameters:
 * * display = Pointer to the display
 * * width   = Width of the display
 * * height  = Height of the display
 * * format  = Pixel format of the frame buffer
 * * printErrors = Whether to print errors (0 = No, 1 = Yes)
 */
int write_display_to_output(INTU8 *display, INTU32 width,
                            INTU32 height, INTU16 format, INTU8 printErrors)
{
	/* Variable Declarations */
	INTU32 displaySize; /* Size of the display (in bytes) */
	size_t fwriteval; /* Value returned from fwrite */
	int    fseekval;  /* Value returned from fseek */

	/* Calculates numPixels and PixelSize */
	displaySize = (INTU32) (format & 0x00FF);
	displaySize >>= 3U;
	displaySize *= width;
	displaySize *= height;

	/* Moves to the top-left corner of the screen */
	fseekval = fseek(framebuffer, 0l, SEEK_SET);
	if (fseekval != 0)
	{
		if (printErrors == (INTU8) 0x01)
		{
			printf(PROGNAME ": Could not move to the top of " FBDEV "\n");
		}
		return ERROR_OSDEP;
	}

	/* Writes the display to the screen */
	fwriteval = fwrite(display, (size_t) 1, (size_t) displaySize, framebuffer);
	if (fwriteval != (size_t) displaySize)
	{
		if (printErrors == (INTU8) 0x01)
		{
			printf(PROGNAME ": Could not write to " FBDEV "\n");
		}
		return ERROR_OSDEP;
	}

	/* Ends the function, as everything went well */
	return ERROR_NONE;
}

